import argparse
import curses
import datetime
import locale
import os
import signal
import time
from concurrent import futures

import blessings

from .version import __version__
from .window import Window
from .common import (
    APP_BG,
    APP_FG,
    APP_TITLE,
    BLK,
    COLORS,
    LOOP_MS,
    MAX_WINS,
    MAX_COLS,
    APP_FRAME,
    APP_CLOCK,
    tprint,
    tformat,
)

TERM_TITLE = "logview"
APPLICATION_TITLE = f"logview v{__version__}"

locale.setlocale(locale.LC_ALL, "")

config = None
windows = []
colors = []
log = []


class Logger:
    @staticmethod
    def append(value):
        log.append(value)


class SignalHook:
    """
    Hooks to SIGINT, SIGTERM, SIGKILL
    """

    SIGINT = signal.SIGINT
    SIGTERM = signal.SIGTERM
    SIGKILL = signal.SIGKILL

    def __init__(self):
        self._last_signal = None
        try:
            signal.signal(signal.SIGINT, self._signal_received)
        except OSError:
            pass
        try:
            signal.signal(signal.SIGTERM, self._signal_received)
        except OSError:
            pass
        try:
            signal.signal(signal.SIGKILL, self._signal_received)
        except OSError:
            pass

    def _signal_received(self, signum, frame):
        self._last_signal = signal.Signals(signum)  # pylint: disable=no-member

    @property
    def signal(self):
        return self._last_signal

    @property
    def exit(self):
        return self._last_signal in [self.SIGINT, self.SIGTERM, self.SIGKILL]


def parse_args():
    parser = argparse.ArgumentParser(
        prog="logview", description="logview allows you to tail multiple files at once"
    )
    parser.add_argument(
        "-t",
        "--trim-date",
        action="store_true",
        help="display only time from timestamp",
    )
    parser.add_argument(
        "-n",
        "--lines",
        metavar="NUM",
        type=int,
        default=1000,
        help="scrollback buffer lines (default 1000)",
    )
    parser.add_argument(
        "-d",
        "--dir",
        default="",
        help="directory to find FILE(s) in",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
        help="show version number and exit",
    )
    parser.add_argument(
        "file", metavar="FILE", nargs="+", help=f"file(s) to tail, maximum {MAX_WINS}"
    )
    return parser.parse_args()


def draw_app(term):
    top_before_title = int((term.width - len(APPLICATION_TITLE)) / 2)
    top_after_title = int(
        ((term.width - len(APPLICATION_TITLE)) / 2)
        + (((term.width - len(APPLICATION_TITLE)) / 2) % 1)
        - 12  # clock space
    )
    tprint(
        term,
        term.move(0, 0),
        term.color(colors[APP_FRAME]),
        term.on_color(colors[APP_BG]),
        BLK * top_before_title,
        term.color(colors[APP_TITLE]),
        term.on_color(colors[APP_FRAME]),
        term.bold,
        APPLICATION_TITLE,
        term.color(colors[APP_FRAME]),
        term.on_color(colors[APP_BG]),
        BLK * top_after_title,
        term.color(colors[APP_CLOCK]),
        term.on_color(colors[APP_FRAME]),
        term.bold,
        datetime.datetime.now().strftime("%H:%M:%S"),
        term.color(colors[APP_FRAME]),
        term.on_color(colors[APP_BG]),
        BLK * 4,
        term.move(term.height - 1, 0),
        # BLK * term.width,
        term.color(colors[APP_CLOCK]),
        term.on_color(colors[APP_FRAME]),
        "TAB: Next Window   SPACE: Jump to end   Q: Quit".center(term.width),
    )
    tformat(term.color(colors[APP_FRAME]), term.on_color(colors[APP_BG]))
    for row in range(1, term.height - 1):
        print(term.move(row, 0) + BLK + term.move(row, term.width - 1) + BLK)
    tprint(term)


def draw_clock(term):
    tprint(
        term,
        term.move(0, term.width - 12),
        term.color(colors[APP_CLOCK]),
        term.on_color(colors[APP_FRAME]),
        term.bold,
        datetime.datetime.now().strftime("%H:%M:%S"),
    )


def calculate_windows(term, update=True):
    col_count = 1
    row_count = 1
    row_limit = 3
    while len(windows) / col_count / row_count > 1:
        if row_count == row_limit:
            row_count = 1
            row_limit += 1
            col_count += 1 if col_count < MAX_COLS else 0
        else:
            row_count += 1

    win_width = int((term.width - 2) / col_count)
    for col in range(col_count):
        col_windows = [i for i in range(len(windows))][
            row_count * col : row_count * (col + 1)
        ]
        win_height = int((term.height - 2) / len(col_windows))
        for win in col_windows:
            windows[win].resize(
                x=win_width * col + 1,
                y=win_height * col_windows.index(win) + 1,
                w=win_width,
                h=win_height,
                update=update,
            )


def main_loop(term, screen):
    sighook = SignalHook()
    selected_window = 0
    windows[selected_window].selected = True
    windows[selected_window].draw_frame()
    while not sighook.exit:
        force = False
        key_code = screen.getch()

        while key_code != curses.ERR:
            force = force or (key_code != curses.ERR)
            if key_code in [ord("q"), ord("Q")]:
                return
            elif key_code == curses.KEY_RESIZE:
                draw_app(term)
                calculate_windows(term)
            elif key_code == 9:  # TAB
                windows[selected_window].selected = False
                selected_window += 1
                if selected_window >= len(windows):
                    selected_window = 0
                windows[selected_window].selected = True
            elif key_code == curses.KEY_UP:  # also scroll up
                windows[selected_window].scroll_up()
            elif key_code == curses.KEY_PPAGE:  # PGUP
                windows[selected_window].page_up()
            elif key_code == curses.KEY_DOWN:  # also scroll down
                windows[selected_window].scroll_down()
            elif key_code == curses.KEY_NPAGE:  # PGDN
                windows[selected_window].page_down()
            elif key_code == ord(" "):
                windows[selected_window].scroll_end()
            key_code = screen.getch()

        draw_clock(term)
        for window in windows:
            window.refresh(force)
        tprint(term)
        time.sleep(LOOP_MS / 1000)


def logview():
    term = blessings.Terminal()
    if not term.is_a_tty:
        print("logview output cannot be piped, it is for display only")
    elif term.number_of_colors < 16:
        print("logview requires a terminal that supports a minimum of 16 colours")
    else:
        global config
        config = parse_args()
        print(f"\x1b]2;{TERM_TITLE}\x07", end="")
        colors.extend(COLORS[term.number_of_colors])
        for filename in config.file:
            filepath = os.path.abspath(os.path.join(config.dir, filename))
            if os.access(filepath, os.R_OK, effective_ids=True):
                windows.append(Window(filepath, config, Logger))
            elif os.access(filepath, os.F_OK, effective_ids=True):
                print(f"File '{filepath}' is not readable")
            else:
                print(f"File '{filepath}' does not exist")
            if windows and not windows[-1].reader.isOpen:
                windows.pop(len(windows) - 1)

        if windows:
            screen = None
            try:
                screen = curses.initscr()
                curses.noecho()
                curses.cbreak()
                screen.keypad(True)
                screen.nodelay(True)
                screen.getch()

                print(
                    term.enter_fullscreen
                    + term.hide_cursor
                    + term.color(colors[APP_FG])
                    + term.on_color(colors[APP_BG])
                    + term.clear
                )
                draw_app(term)
                calculate_windows(term, False)
                for window in windows:
                    window.start(term)
                    # window.load()
                    # window.refresh(True)
                with futures.ThreadPoolExecutor() as pool:
                    threads = {pool.submit(window.load): window for window in windows}
                    for future in futures.as_completed(threads):
                        threads[future].refresh(True)
                main_loop(term, screen)
            except Exception as err:
                raise err
            finally:
                tprint(term, term.exit_fullscreen + term.normal_cursor)
                curses.nocbreak()
                if screen:
                    screen.keypad(False)
                curses.echo()
                curses.endwin()
                print(*log, sep="\n")
            exit(0)
    exit(1)


if __name__ == "__main__":
    logview()



__all__ = ["LogRecord"]

import typing as t
import re
import datefinder
import textwrap
from .common import DELIMETER_PATTERNS, LOG_BG, LOG_FG, LOG_FG_DARK, LOG_LEVEL

if t.TYPE_CHECKING:
    from .window import Window


class Line:
    __slots__ = ["_text", "_format_pre", "_format_post"]

    def __init__(self, text, format_pre, format_post):
        self._text = text
        self._format_pre = format_pre
        self._format_post = format_post

    @property
    def text(self) -> str:
        return self._text

    @property
    def format_pre(self) -> str:
        return self._format_pre

    @property
    def format_post(self) -> str:
        return self._format_post

    @property
    def length(self) -> int:
        return len(self.text)

    @property
    def formatted(self) -> str:
        return self.get_formatted()[0]

    def get_formatted(self, width=0) -> t.List[str]:
        if not width or self.length <= width:
            return ["".join([self._format_pre, self._text, self._format_post])]

        return [
            "".join([self._format_pre, part, self._format_post])
            for part in textwrap.wrap(
                self._text,
                width=width,
                replace_whitespace=False,
                drop_whitespace=False,
            )
        ]


class Header(Line):
    __slots__ = ["_parts"]

    def __init__(self, window, text):
        self._parts: t.List[Line] = []
        self._parse(text, window)

    @property
    def text(self) -> str:
        return "".join([part.text for part in self._parts])

    @property
    def format_pre(self) -> str:
        return self._parts[-1]._format_pre

    @property
    def format_post(self) -> str:
        return self._parts[-1]._format_post

    def get_formatted(self, width=0) -> t.List[str]:
        if not width or self.length < width:
            return ["".join([part.formatted for part in self._parts])]

        lines = textwrap.wrap(
            self.text, width, replace_whitespace=False, drop_whitespace=False
        )
        display_lines: t.List[str] = [""] * len(lines)
        part_index = 0
        part_seek_pos = 0

        for line_index in range(len(lines)):
            line = lines[line_index]
            # if part_index==len(self._parts)-1 and not display_lines[line_index]:
            #    display_lines[line_index]="  "
            while line:
                part = self._parts[part_index]
                line_trim = 0
                if len(part.text[part_seek_pos:]) <= len(line):
                    display_lines[line_index] += (
                        f"{part._format_pre}"
                        f"{part.text[part_seek_pos:]}"
                        f"{part.format_post}"
                    )
                    line_trim = len(part.text[part_seek_pos:])
                    part_index += 1
                    part_seek_pos = 0
                else:
                    display_lines[line_index] += (
                        f"{part._format_pre}"
                        f"{part.text[part_seek_pos:part_seek_pos + len(line)]}"
                        f"{part.format_post}"
                    )
                    line_trim = len(
                        part.text[part_seek_pos : part_seek_pos + len(line)]
                    )
                    part_seek_pos += len(
                        part.text[part_seek_pos : part_seek_pos + len(line)]
                    )
                line = line[line_trim:]

        return display_lines

    def _parse(self, text, window):
        FORMAT_LOG_BG = window.term.on_color(window.colors[LOG_BG])
        FORMAT_LOG_FG = window.term.color(window.colors[LOG_FG])
        FORMAT_LOG_FG_DARK = window.term.color(window.colors[LOG_FG_DARK])

        if window.config.trim_date:
            # replace timestamp with HH:MM:SS.FFF
            timestamp_match = datefinder.find_dates(text, source=True, index=True)
            timestamp = next(timestamp_match)
            if timestamp:
                timestamp_time = timestamp[0].strftime("%H:%M:%S.%f")[:-3]
                timestamp_end = timestamp[2][1]
                timestamp_start = timestamp_end - len(timestamp[1]) - 1
                text = (
                    f"{text[:timestamp_start]}{timestamp_time}{text[timestamp_end-1:]}"
                )

        # find log level
        text_upper = text.upper()
        text_len = len(text)
        level_matches = [text_upper.find(key) for key in LOG_LEVEL.keys()]
        level_start = min(level_matches, key=lambda v: text_len if v == -1 else v)
        del text_upper, text_len
        try:
            level_key = list(LOG_LEVEL.keys())[level_matches.index(level_start)]
            level_color = window.term.color(window.colors[LOG_LEVEL[level_key]])
        except IndexError:
            level_key = ""
            level_start = 0
            level_color = window.term.color(window.colors[LOG_FG])

        # find delimiter for end of header / start of message
        delim_end = 0
        for pattern in DELIMETER_PATTERNS:
            match = re.match(pattern, text)
            if match is not None:
                delim_end = match.end() - 1
                break

        # build header
        self._parts.append(
            Line(
                text=text[:level_start],
                format_pre=f"{FORMAT_LOG_BG}{FORMAT_LOG_FG_DARK}",
                format_post=f"{FORMAT_LOG_FG}",
            )
        )
        self._parts.append(
            Line(
                text=text[level_start : level_start + len(level_key)],
                format_pre=f"{FORMAT_LOG_BG}{window.term.bold}{level_color}",
                format_post=f"{window.term.normal}{FORMAT_LOG_BG}",
            )
        )
        self._parts.append(
            Line(
                text=text[level_start + len(level_key) : delim_end],
                format_pre=f"{FORMAT_LOG_BG}{FORMAT_LOG_FG_DARK}",
                format_post=f"{FORMAT_LOG_FG}",
            )
        )

        # add remaining text
        self._parts.append(
            Line(
                text=text[delim_end:],
                format_pre=f"{FORMAT_LOG_BG}{level_color}",
                format_post=f"{FORMAT_LOG_FG}",
            )
        )


class LogRecord:
    __slots__ = ["_window", "_header", "_lines", "_width", "_display_lines"]

    def __init__(self, window: "Window", lines: t.List[str]):
        self._window = window
        self._lines: t.List[Line] = []
        self._width = 0
        self._display_lines: t.List[str] = []

        if not lines[-1]:
            # trim blank last line, this is from the pattern matcher
            lines = lines[:-1]
        self._lines.append(Header(window, lines[0]))
        lines.pop(0)
        for line in lines:
            self._lines.append(
                Line(line, self._lines[0].format_pre, self._lines[0].format_post)
            )

        self._update_display_lines()

    @property
    def display_lines(self) -> t.List[str]:
        """Return a list of strings with format codes, wrapped for the window """
        if self._width != self._window.w - 2:
            self._update_display_lines()
        return self._display_lines

    def append(self, lines: t.List[str]):
        """Add more lines to this record"""
        for line in lines:
            self._lines.append(
                Line(
                    text=line,
                    format_pre=self._lines[-1].format_pre,
                    format_post=self._lines[-1].format_post,
                )
            )
        self._update_display_lines()

    ##### Internal Methods #####

    def _update_display_lines(self):
        width = self._window.w - 2
        self._width = width
        self._display_lines: t.List[str] = []
        for line in self._lines:
            self._display_lines.extend(line.get_formatted(width))



import typing as t
import datefinder
from .record import LogRecord

if t.TYPE_CHECKING:
    from .window import Window


class LogReader:
    __slots__ = [
        "_window",
        "filename",
        "_lines",
        "records",
        "_record_prefix_length",
        "_fd",
        "_buffer",
    ]

    def __init__(self, window: "Window", filename: str):
        self._window = window
        self.filename = filename
        self._lines = 0  # file line count
        self.records: t.List[LogRecord] = []
        self._record_prefix_length: t.Union[
            str, None
        ] = None  # pattern that starts record
        self._fd = None
        self._buffer = bytes()  # byte buffer for file data
        self._open()

    def __del__(self):
        if self._fd:
            try:
                self._fd.close()
            except:
                pass

    @property
    def isOpen(self) -> bool:
        if self._fd is not None:
            return not self._fd.closed
        return False

    @property
    def lines(self) -> int:
        return self._lines

    def preload(self) -> t.List[LogRecord]:
        self._get_line_count()
        self._read_last()
        return self.read(0)

    def read(self, records: int = 1) -> t.List[LogRecord]:
        """Read to end of file and parse next line"""
        self._read()
        return self._get_records(records)

    ##### Internal Methods #####

    def _open(self):
        try:
            self._fd = open(self.filename, "rb")
        except IOError as err:
            self._window.log.append(
                f"Failed to open '{self.filename}': [{err.errno}] {err.strerror}"
            )
            self._fd = None

    def _read(self):
        if self.isOpen:
            try:
                data = self._fd.read()
                self._lines += data.count(b"\n")
                self._buffer += data
            except IOError as err:
                self._window.log.append(
                    f"Error while reading '{self.filename}': [{err.errno}] {err.strerror}"
                )

    def _get_line_count(self):
        if not self.isOpen:
            return
        try:
            self._fd.seek(0, 0)
            self._lines = 0
            for line in self._fd:  # pylint: disable=unused-variable
                self._lines += 1
        except IOError as err:
            self._fd = None
            self._lines = 0
            self._window.log.append(
                f"Error while reading '{self.filename}': [{err.errno}] {err.strerror}"
            )

    def _read_last(self):
        """Read last ``lines`` of file, like 'tail -n'"""
        if not self.isOpen:
            return
        try:
            last_read_block = self._fd.tell()
            block_end_byte = last_read_block
            BLOCK_SIZE = min(block_end_byte, 1024)
            remain_lines = self._window.config.lines
            block_num = -1
            blocks = []
            while remain_lines > 0 and block_end_byte > 0:
                if block_end_byte - BLOCK_SIZE > 0:
                    self._fd.seek(block_num * BLOCK_SIZE, 2)
                    blocks.append(self._fd.read(BLOCK_SIZE))
                else:
                    self._fd.seek(0, 0)
                    blocks.append(self._fd.read(block_end_byte))
                remain_lines -= blocks[-1].count(b"\n")
                block_end_byte -= BLOCK_SIZE
                block_num -= 1
            self._fd.seek(last_read_block, 0)
        except IOError as err:
            self._fd = None
            self._window.log.append(
                f"Error while reading '{self.filename}': [{err.errno}] {err.strerror}"
            )
        else:
            for block in blocks[::-1]:
                self._buffer += block

    def _find_record_prefix_length(self):
        """
        Rudamentary prefix length finder. Looks for repeated same number of chars
        between newline/file-start and timestamp.
        """
        prefix_lengths = []
        last_end = 1
        buffer_string = self._buffer.decode()
        for result in datefinder.find_dates(buffer_string, source=True, index=True):
            if self._record_prefix_length is not None:
                break
            elif len(result[1]) < 6:
                # skip matches too short, probably just numbers not a timestamp
                continue
            timestamp_end = result[2][1]
            timestamp_start = timestamp_end - len(result[1]) - 1
            prefix_lengths.append(
                len(
                    buffer_string[
                        timestamp_start
                        - buffer_string[last_end:timestamp_start][::-1].find(
                            "\n"
                        ) : timestamp_start
                    ]
                )
            )
            last_end = buffer_string.find("\n", timestamp_end)
            for length in prefix_lengths:
                if prefix_lengths.count(length) > 3:
                    self._record_prefix_length = length
                    break

    def _get_records(self, count) -> t.List[LogRecord]:
        # make sure we have a lock on the prefix before parsing anything
        if self._record_prefix_length is None:
            self._find_record_prefix_length()
            if self._record_prefix_length is None:
                return []

        # find timestamps in buffer
        buffer_string = self._buffer.decode()
        timestamp_match = datefinder.find_dates(buffer_string, source=True, index=True)
        this_result = next(timestamp_match, None)
        match_offset = 0
        new_records = 0
        while True:
            if len(self._buffer) == 0:
                break

            if this_result is None:
                # if no timestamps found, read to the end of the buffer
                this_start = len(self._buffer)
            else:
                # skip if this match is too short, sometimes numbers will match
                if len(this_result[1]) < 8:
                    this_result = next(timestamp_match, None)
                    continue
                # this should always work out to 0, anything else means there is
                # data in the buffer that belongs to the previous record
                this_start = (
                    this_result[2][1]
                    - match_offset
                    - len(this_result[1])
                    - 1
                    - self._record_prefix_length
                )

            if this_start != 0:
                # make sure this match is a record
                if this_result is not None and buffer_string[this_start - 1] != "\n":
                    this_result = next(timestamp_match, None)
                    continue
                # something left in the buffer before the this record start
                # append it to the last record and search again
                if self.records:
                    # only keep if there is at least one record, anything else is
                    # a partial record from tail-ing the file
                    self.records[-1].append(buffer_string[:this_start].split("\n"))
                self._buffer = self._buffer[len(buffer_string[:this_start].encode()) :]
                buffer_string = buffer_string[this_start:]
                match_offset += this_start
                this_start -= match_offset

            while True:
                next_result = next(timestamp_match, None)
                if next_result is None:
                    # no second timestamp found, read to the end of the buffer
                    next_start = len(self._buffer)
                else:
                    # if len(next_result[1]) < 8:
                    #    continue
                    next_start = (
                        next_result[2][1]
                        - match_offset
                        - len(next_result[1])
                        - 1
                        - self._record_prefix_length
                    )
                if next_result is None or buffer_string[next_start - 1] == "\n":
                    self.records.append(
                        LogRecord(
                            self._window,
                            buffer_string[:next_start].split("\n"),
                        )
                    )
                    self._buffer = self._buffer[
                        len(buffer_string[:next_start].encode()) :
                    ]
                    buffer_string = buffer_string[next_start:]
                    match_offset += next_start
                    new_records += 1
                    break

            if next_result is None or new_records == count:
                break
            else:
                this_result = next_result

        self.records = self.records[0 - self._window.config.lines :]
        return self.records[0 - new_records :] if new_records else []



__all__ = ["Window"]

import typing as t
import os
import locale
from .common import (
    ACTIVE_DELAY,
    BLK,
    BLK_B,
    BLK_BL,
    BLK_BR,
    BLK_L,
    BLK_R,
    COLORS,
    LOG_BG,
    WIN_FRAME,
    WIN_FRAME_ACTIVE,
    WIN_FRAME_ERROR,
    WIN_FRAME_LOAD,
    WIN_FRAME_SELECT,
    WIN_FRAME_SELECT_ACTIVE,
    WIN_LINES,
    WIN_TITLE,
    tprint,
    tformat,
)
from .reader import LogReader

if t.TYPE_CHECKING:
    from .record import LogRecord


class Window(object):
    __slots__ = [
        "config",
        "log",
        "colors",
        "name",
        "term",
        "_selected",
        "active_delay",
        "reader",
        "_display_lines",
        "_last_line",
        "frame",
        "x",
        "y",
        "w",
        "h",
    ]

    def __init__(self, filename, config, log):
        self.config = config
        self.log = log
        self.colors = []
        self.name = os.path.split(filename)[-1]
        self.term = None
        self._selected = False
        self.active_delay = 0
        self.reader = LogReader(self, filename)
        self._display_lines = []
        self._last_line = 0
        self.frame = None
        self.x = 0
        self.y = 0
        self.w = 0
        self.h = 0

    @property
    def selected(self):
        return self._selected

    @selected.setter
    def selected(self, value):
        self._selected = value
        if value:
            self.frame = self.colors[
                WIN_FRAME_SELECT_ACTIVE if self.active_delay else WIN_FRAME_SELECT
            ]
        else:
            self.frame = self.colors[
                WIN_FRAME_ACTIVE if self.active_delay else WIN_FRAME
            ]

    def start(self, term):
        self.term = term
        self.colors.extend(COLORS[term.number_of_colors])
        self.frame = self.colors[WIN_FRAME_LOAD]
        self.draw_frame(True)

    def load(self):
        self.reader.preload()
        self._update_display_lines()
        if self.reader.isOpen:
            self.frame = self.colors[WIN_FRAME_SELECT if self.selected else WIN_FRAME]
        else:
            self.frame = self.colors[WIN_FRAME_ERROR]

    def resize(self, x, y, w, h, update):
        self.x = x
        self.y = y
        self.w = w
        self.h = h
        if update:
            self._update_display_lines()
            self.refresh(True)

    def refresh(self, force=False):
        new_records = self.reader.read(0)
        if new_records and not force:
            self.active_delay = ACTIVE_DELAY
            self.frame = self.colors[
                WIN_FRAME_SELECT_ACTIVE if self._selected else WIN_FRAME_ACTIVE
            ]
            self._update_display_lines(new_records)

        if new_records or force:
            self.draw_frame()
            tformat(self.term.on_color(self.colors[LOG_BG]))
            for i in range(self.h - 2):
                print(
                    self.term.move(self.y + (self.h - 2 - i), self.x + 1)
                    + (" " * (self.w - 2))
                    + self.term.move(self.y + (self.h - 2 - i), self.x + 1)
                    + (
                        self._display_lines[
                            len(self._display_lines) + self._last_line - i - 1
                        ]
                        if len(self._display_lines) + self._last_line - i - 1 >= 0
                        else (" " * (self.w - 2))
                    )
                )

        if self._last_line == 0 and not new_records:
            if self.active_delay > 0:
                self.active_delay -= 1
            else:
                new_frame = self.colors[
                    WIN_FRAME_SELECT if self._selected else WIN_FRAME
                ]
                if self.frame != new_frame:
                    self.frame = new_frame
                    self.draw_frame()
                else:
                    self.frame = new_frame

    def scroll_up(self, lines=1):
        if len(self._display_lines) < self.h - 2:
            return
        max_scroll = 0 - len(self._display_lines) + 1 + (self.h - 2)
        if self._last_line > max_scroll:
            self._last_line -= lines
        if self._last_line < max_scroll:
            self._last_line = max_scroll

    def scroll_down(self, lines=1):
        if self._last_line < 0:
            self._last_line += lines
        if self._last_line > 0:
            self._last_line = 0

    def scroll_end(self):
        self._last_line = 0

    def page_up(self):
        self.scroll_up(self.h - 2)

    def page_down(self):
        self.scroll_down(self.h - 2)

    def draw_frame(self, fill=False):
        tprint(  # draw top edge and corners
            self.term,
            self.term.move(self.y, self.x),
            self.term.color(self.frame) + self.term.on_color(self.colors[LOG_BG]),
            BLK * int((self.w - len(self.name)) / 2),
            self.term.color(self.colors[WIN_TITLE]) + self.term.on_color(self.frame),
            self.term.bold,
            self.name,
            self.term.color(self.frame) + self.term.on_color(self.colors[LOG_BG]),
            BLK
            * int(
                (
                    ((self.w - len(self.name)) / 2)
                    + (((self.w - len(self.name)) / 2) % 1)
                )
                - 18
            ),
            self.term.color(self.colors[WIN_LINES]),
            self.term.on_color(self.frame),
            self.term.bold,
            f"lines: {locale.format_string('%d', self.reader.lines, True):>9}",
            self.term.color(self.frame) + self.term.on_color(self.colors[LOG_BG]),
            BLK * 2,
        )
        tprint(  # draw bottom edge and corners
            self.term,
            self.term.move(self.y + self.h - 1, self.x),
            self.term.color(self.frame) + self.term.on_color(self.colors[LOG_BG]),
            BLK_BL,
            BLK_B * (self.w - 2),
            BLK_BR,
        )
        # draw left and right edge and fill window
        tformat(self.term.color(self.frame) + self.term.on_color(self.colors[LOG_BG]))
        if fill:
            for row in range(self.y + 1, self.y + self.h - 1):
                print(
                    self.term.move(row, self.x) + BLK_L + (" " * (self.w - 2)) + BLK_R
                )
        else:
            for row in range(self.y + 1, self.y + self.h - 1):
                print(
                    self.term.move(row, self.x)
                    + BLK_L
                    + self.term.move(row, self.x + self.w - 1)
                    + BLK_R
                )
        # reset formatting
        tprint(self.term)

    def _update_display_lines(self, new_records: "t.List[LogRecord]" = []):
        if new_records:
            scroll_at_end = self._last_line == 0
            lines_added = 0
            for record in new_records:
                new_lines = record.display_lines
                lines_added += len(new_lines)
                self._display_lines.extend(new_lines)

            if not scroll_at_end:
                self.scroll_up(len(self._display_lines) - lines_added)
            self._display_lines[
                sum([len(record.display_lines) for record in new_records]) :
            ]
        else:
            self._display_lines = []
            for record in self.reader.records:
                self._display_lines.extend(record.display_lines)
            self._last_line = 0



DELIMETER_PATTERNS = [r".*( \| )", r".*( - )", r".*(\] [^\[])"]

MAX_COLS = 3
MAX_ROWS = 4
MAX_WINS = MAX_COLS * MAX_ROWS
ACTIVE_DELAY = 10
LOOP_MS = 50

BLK_T = "\u2580"
BLK_B = "\u2584"
BLK = "\u2588"
BLK_L = "\u258C"
BLK_R = "\u2590"
BLK_BL = "\u2599"
BLK_TL = "\u259B"
BLK_TR = "\u259C"
BLK_BR = "\u259F"

APP_FG = 0
APP_BG = 1
APP_FRAME = 2
APP_TITLE = 3
APP_CLOCK = 4
WIN_FRAME = 5
WIN_FRAME_ACTIVE = 6
WIN_FRAME_SELECT = 7
WIN_FRAME_SELECT_ACTIVE = 8
WIN_FRAME_LOAD = 9
WIN_FRAME_ERROR = 10
WIN_TITLE = 11
WIN_LINES = 12
LOG_FG = 13
LOG_FG_DARK = 14
LOG_BG = 15
LOG_LEVEL = {
    "TRACE": 16,
    "DEBUG": 17,
    "INFO": 18,
    "WARNING": 19,
    "WARN": 19,
    "ERROR": 20,
    "CRITICAL": 21,
    "EMERG": 21,
}
# fmt: off
COLORS = {
    16: [
        15, # APP_FG
        16, # APP_BG
        4,  # APP_FRAME
        3,  # APP_TITLE
        16, # APP_CLOCK
        2,  # WIN_FRAME
        10, # WIN_FRAME_ACTIVE
        6,  # WIN_FRAME_SELECT
        14, # WIN_FRAME_SELECT_ACTIVE
        3,  # WIN_FRAME_LOAD
        1,  # WIN_FRAME_ERROR
        16, # WIN_TITLE
        16, # WIN_LINES
        15, # LOG_FG
        8,  # LOG_FG_DARK
        16, # LOG_BG
        8,  # LOG_LEVEL: TRACE
        8,  # LOG_LEVEL: DEBUG
        10, # LOG_LEVEL: INFO
        11, # LOG_LEVEL: WARN
        9,  # LOG_LEVEL: ERROR
        1,  # LOG_LEVEL: CRITICAL
    ],
    88: [
        15, # APP_FG
        16, # APP_BG
        4,  # APP_FRAME
        3,  # APP_TITLE
        16, # APP_CLOCK
        28, # WIN_FRAME
        46, # WIN_FRAME_ACTIVE
        6,  # WIN_FRAME_SELECT
        45, # WIN_FRAME_SELECT_ACTIVE
        3,  # WIN_FRAME_LOAD
        1,  # WIN_FRAME_ERROR
        16, # WIN_TITLE
        16, # WIN_LINES
        15, # LOG_FG
        8,  # LOG_FG_DARK
        16, # LOG_BG
        8,  # LOG_LEVEL: TRACE
        8,  # LOG_LEVEL: DEBUG
        40, # LOG_LEVEL: INFO
        11, # LOG_LEVEL: WARN
        9,  # LOG_LEVEL: ERROR
        52, # LOG_LEVEL: CRITICAL
    ],
    256: [
        15, # APP_FG
        16, # APP_BG
        4,  # APP_FRAME
        3,  # APP_TITLE
        16, # APP_CLOCK
        28, # WIN_FRAME
        46, # WIN_FRAME_ACTIVE
        6,  # WIN_FRAME_SELECT
        45, # WIN_FRAME_SELECT_ACTIVE
        3,  # WIN_FRAME_LOAD
        1,  # WIN_FRAME_ERROR
        16, # WIN_TITLE
        16, # WIN_LINES
        15, # LOG_FG
        240,# LOG_FG_DARK
        233,# LOG_BG
        242,# LOG_LEVEL: TRACE
        246,# LOG_LEVEL: DEBUG
        40, # LOG_LEVEL: INFO
        208,# LOG_LEVEL: WARN
        9,  # LOG_LEVEL: ERROR
        124,# LOG_LEVEL: CRITICAL
    ]
}
# fmt: on


def tprint(term, *s):
    if len(s) == 0:
        print(term.move(0, 0) + term.normal)
    else:
        print(*s, sep="", end=term.normal)


def tformat(*s):
    print(sep="", end="".join(s))



import setuptools
import os

with open("README.md") as fd:
    long_description = fd.read()


def get_version():
    with open(
        os.path.join(
            os.path.abspath(os.path.dirname(__file__)), "pylogview", "version.py"
        )
    ) as fd:
        for line in fd.readlines():
            if line.startswith("__version__"):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
        else:
            raise RuntimeError("Unable to find version string.")


setuptools.setup(
    name="pylogview",
    version=get_version(),
    author="Michael Murton",
    description=(
        "logview is a simple terminal script that allows you to tail multiple log "
        "files in a windowed format."
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/CrazyIvan359/logview",
    packages=["pylogview"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Environment :: Console :: Curses",
        # "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Topic :: Home Automation",
        "Topic :: Internet :: Log Analysis",
        "Topic :: System :: Logging",
        "Topic :: Utilities",
    ],
    license="MIT",
    keywords="log, logging, tail",
    install_requires=["blessings>=1.7,<2.0"],
    entry_points={"console_scripts": ["logview=pylogview.__main__:main"]},
    zip_safe=False,
)
